"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = __importDefault(require("events"));
const net_1 = __importDefault(require("net"));
const crypto_1 = __importDefault(require("crypto"));
const bsv_minimal_1 = require("bsv-minimal");
const messages_1 = require("./messages");
const config_1 = require("./config");
const events_2 = __importDefault(require("./events"));
class Peer extends events_1.default {
    constructor({ node, port, ticker = "BSV", stream = true, validate = true, autoReconnect = true, autoReconnectWait = 1000 * 2, // 2 seconds
    segwit = config_1.SEGWIT[ticker] || config_1.SEGWIT.DEFAULT, disableExtmsg = false, DEBUG_LOG = false, magic = config_1.MAGIC_NUMS[ticker] || config_1.MAGIC_NUMS.DEFAULT, version = config_1.VERSIONS[ticker] || config_1.VERSIONS.DEFAULT, user_agent, start_height = 0, mempoolTxs = true, }) {
        super();
        this.setMaxListeners(0);
        this.magic = magic;
        this.version = version;
        this.user_agent = user_agent;
        this.start_height = start_height;
        this.mempoolTxs = mempoolTxs;
        this.segwit = segwit;
        this.port = port || 8333;
        if (!port && node.split(":").length > 1) {
            const split = node.split(":");
            const portNum = parseInt(split[split.length - 1]);
            if (portNum > 0)
                this.port = portNum;
        }
        this.node = node;
        if (node.split(":").length === 2) {
            this.node = node.split(":")[0];
        }
        else if (node.includes("[") && node.split("]").length === 2) {
            // ipv6 formated with port https://en.wikipedia.org/wiki/IPv6#Addressing
            this.node = node.replace("[", "").split("]")[0];
        }
        this.ticker = ticker;
        this.stream = stream;
        this.validate = validate;
        this.autoReconnect = autoReconnect;
        this.autoReconnectWait = autoReconnectWait;
        this.disableExtmsg = disableExtmsg;
        this.connected = false;
        this.extmsg = false;
        this.disconnects = 0;
        this.timeoutConnect = 1000 * 30; // 30 seconds
        this.DEBUG_LOG = DEBUG_LOG;
        this.emitter = new events_2.default();
        this.buffers = {
            data: [],
            needed: 0,
            length: 0,
            chunkNum: 0,
            block: new bsv_minimal_1.Block({ validate }),
            downloadingBlock: false,
        };
    }
    sendMessage(command, payload, force = false) {
        if (!this.connected && !force)
            throw Error(`Not connected`);
        const { magic, extmsg } = this;
        const serialized = messages_1.Message.write({
            command,
            payload,
            magic,
            extmsg,
        });
        if (!this.socket)
            throw Error("Socket not connected");
        this.socket.write(serialized);
        this.DEBUG_LOG &&
            console.log(`bsv-p2p: Sent message ${command} ${payload ? payload.length : "0"} bytes`);
    }
    streamBlock(chunk) {
        const { buffers, ticker, validate, node, port } = this;
        const { finished, started, transactions, bytesRemaining, header, height, size, txCount, startDate, } = buffers.block.addBufferChunk(chunk);
        const blockHash = header.getHash();
        this.emitter.resetTimeout(`block_${blockHash.toString("hex")}`, 30); // Extend getBlock timeout another 30 seconds
        if (finished) {
            if (bytesRemaining > 0 && buffers.block.br) {
                const remaining = buffers.block.br.readAll();
                buffers.data = [remaining];
                buffers.length = remaining.length;
            }
            else {
                buffers.data = [];
                buffers.length = 0;
            }
            buffers.block = new bsv_minimal_1.Block({ validate });
            buffers.needed = 0;
            buffers.downloadingBlock = false;
            this.emitter.emit(`block_${blockHash.toString("hex")}`, {
                ticker,
                blockHash,
                header,
                height,
                size,
                startDate,
            });
        }
        this.emit("block_chunk", {
            node,
            port,
            num: buffers.chunkNum++,
            started,
            finished,
            transactions,
            header,
            ticker,
            chunk: finished
                ? chunk.subarray(0, chunk.length - bytesRemaining)
                : chunk,
            blockHash,
            height,
            size,
            txCount,
            startDate,
        });
        this.emit("transactions", {
            node,
            port,
            ticker,
            finished,
            started,
            transactions,
            bytesRemaining,
            header,
            height,
            size,
            txCount,
            startDate,
        });
    }
    readMessage(buffer) {
        const { node, port, magic, buffers, ticker, stream, validate, listenTxs, listenBlocks, extmsg, } = this;
        const message = messages_1.Message.read({ buffer, magic, extmsg });
        const { command, payload, end, needed } = message;
        buffers.needed = needed;
        if (stream && command === "block") {
            buffers.chunkNum = 0;
            buffers.downloadingBlock = true;
            buffers.block = new bsv_minimal_1.Block({ validate });
            if (payload.length > 0) {
                try {
                    this.streamBlock(payload);
                }
                catch (err) {
                    // Not enough data to parse block header and txCount. Wait for more.
                }
            }
        }
        if (needed)
            return;
        const remainingBuffer = buffer.subarray(end);
        buffers.data = [remainingBuffer];
        buffers.length = remainingBuffer.length;
        buffers.needed = 0;
        this.DEBUG_LOG &&
            command !== "inv" &&
            console.log(`bsv-p2p: Received message`, command, payload && `${payload.length} bytes`);
        if (command === "ping") {
            this.sendMessage("pong", payload);
            this.emit("ping", { ticker, node, port });
        }
        else if (command === "pong") {
            const nonce = payload.toString("hex");
            this.emitter.emit(`pong_${nonce}`);
            this.emit("pong", { ticker, node, port, nonce });
        }
        else if (command === "headers") {
            const { headers, txs } = messages_1.Headers.parseHeaders(payload);
            this.DEBUG_LOG &&
                console.log(`bsv-p2p: Received ${headers.length} headers`);
            this.emitter.emit("headers", headers);
            this.emit(`headers`, { ticker, node, port, headers, txs });
        }
        else if (command === "version") {
            this.sendMessage("verack", null, true);
            const version = messages_1.Version.read(payload);
            this.DEBUG_LOG && console.log(`bsv-p2p: version`, version);
            if (!this.disableExtmsg)
                this.extmsg = version.version >= 70016; // Enable/disable extension messages based on node version
            this.emitter.emit("version", { version });
            this.emit("version", { ticker, node, port, version });
        }
        else if (command === "verack") {
            this.DEBUG_LOG && console.log(`bsv-p2p: verack`);
            this.emitter.emit("verack");
        }
        else if (command === "inv") {
            const msg = messages_1.Inv.read(payload);
            this.DEBUG_LOG &&
                console.log(`bsv-p2p: inv`, Object.keys(msg)
                    .filter((k) => msg[k].length > 0)
                    .map((k) => `${k}: ${msg[k].length}`)
                    .join(", "));
            this.emit("inv", msg);
            const { blocks, txs } = msg;
            if (blocks.length > 0) {
                this.emit("block_hashes", { ticker, node, port, hashes: blocks });
            }
            if (listenTxs && txs.length > 0) {
                try {
                    const results = listenTxs(txs);
                    if (results instanceof Promise) {
                        results
                            .then((txids) => this.getTxs(txids))
                            .catch((err) => {
                            this.DEBUG_LOG &&
                                console.error(`bsv-p2p: fetchMempoolTxs threw error: ${err.message}`);
                        });
                    }
                    else {
                        this.getTxs(results);
                    }
                }
                catch (err) {
                    this.DEBUG_LOG &&
                        console.error(`bsv-p2p: fetchMempoolTxs threw error: ${err.message}`);
                }
            }
            if (listenBlocks && blocks.length > 0) {
                try {
                    const results = listenBlocks(blocks);
                    if (results instanceof Promise) {
                        results
                            .then((hashes) => this.getBlocks(hashes))
                            .catch((err) => {
                            this.DEBUG_LOG &&
                                console.error(`bsv-p2p: fetchNewBlocks threw error: ${err.message}`);
                        });
                    }
                    else {
                        this.getBlocks(results);
                    }
                }
                catch (err) {
                    this.DEBUG_LOG &&
                        console.error(`bsv-p2p: fetchNewBlocks threw error: ${err.message}`);
                }
            }
        }
        else if (command === "block") {
            if (!stream) {
                const block = bsv_minimal_1.Block.fromBuffer(payload);
                block.options = { validate };
                this.DEBUG_LOG &&
                    console.log(`bsv-p2p: block`, block.getHash().toString("hex"));
                if (this.listenerCount("transactions") > 0) {
                    block.getTransactionsAsync((params) => {
                        this.emit("transactions", Object.assign(Object.assign({}, params), { ticker, node, port }));
                    });
                }
                const blockHash = block.getHash();
                const { header, size } = block;
                const startDate = +new Date();
                let height;
                try {
                    height = block.getHeight();
                }
                catch (err) { }
                this.emitter.emit(`block_${blockHash.toString("hex")}`, {
                    ticker,
                    blockHash: block.getHash(),
                    header,
                    height,
                    size,
                    startDate,
                });
                this.emit("block", { block, ticker, node, port });
            }
        }
        else if (command === "tx") {
            const transaction = bsv_minimal_1.Transaction.fromBuffer(payload);
            this.DEBUG_LOG && console.log(`bsv-p2p: tx`, transaction);
            this.emit("transactions", {
                ticker,
                node,
                port,
                finished: true,
                transactions: [[0, transaction, 0, transaction.length]],
            });
        }
        else if (command === "notfound") {
            const notfound = messages_1.Inv.read(payload);
            this.DEBUG_LOG && console.log("bsv-p2p: notfound", notfound);
            notfound.blocks.map((hash) => this.emitter.emit(`notfound_block_${hash.toString("hex")}`, "Block not found"));
            this.emit(`notfound`, notfound);
        }
        else if (command === "alert") {
            this.DEBUG_LOG && console.warn(`bsv-p2p: alert ${payload.toString()}`);
            this.emit(`alert`, { ticker, node, port, payload });
        }
        else if (command === "getdata") {
            const msg = messages_1.GetData.read(payload);
            msg.txs.map((hash) => {
                this.emitter.emit(`getdata_tx_${hash.toString("hex")}`);
            });
            this.emit(`getdata`, msg);
        }
        else if (command === "reject") {
            const msg = messages_1.Reject.read(payload);
            this.DEBUG_LOG && console.log(`bsv-p2p: reject`, msg);
            if (msg.data)
                this.emitter.emit(`reject_${msg.data.toString("hex")}`, msg.reason);
            this.emit(`reject`, msg);
        }
        else if (command === "addr") {
            const addrs = messages_1.Address.readAddr(payload);
            this.DEBUG_LOG && console.log(`bsv-p2p: addr`, addrs);
            this.emitter.emit("addr", { ticker, node, port, addrs });
            this.emit("addr", { ticker, node, port, addrs });
        }
        else if (command === "getheaders") {
            this.DEBUG_LOG && console.log(`bsv-p2p: getheaders`);
            this.emit(`getheaders`, { ticker, node, port });
        }
        else if (command === "sendcmpct") {
            this.DEBUG_LOG &&
                console.log(`bsv-p2p: sendcmpct ${payload.toString("hex")}`);
            this.emit(`sendcmpct`, { ticker, node, port, payload });
        }
        else if (command === "sendheaders") {
            this.DEBUG_LOG && console.log(`bsv-p2p: sendheaders`);
            this.emit(`sendheaders`, { ticker, node, port, payload });
        }
        else {
            this.DEBUG_LOG &&
                console.log(`bsv-p2p: Unknown command ${command}, ${payload === null || payload === void 0 ? void 0 : payload.toString("hex")} ${payload === null || payload === void 0 ? void 0 : payload.length} bytes`);
            this.emit(`unknown_msg`, { ticker, node, port, command, payload });
        }
        this.emit("message", { ticker, node, port, command, payload });
        if (remainingBuffer.length > 0) {
            this.readMessage(remainingBuffer);
        }
    }
    connect(options = this.connectOptions) {
        if (!this.promiseConnect) {
            this.promiseConnect = new Promise((resolve, reject) => {
                this.connectOptions = options;
                this.socket = new net_1.default.Socket();
                const { socket, buffers, ticker, version, user_agent, start_height, mempoolTxs, node, port, } = this;
                this.DEBUG_LOG &&
                    console.log(`bsv-p2p: Connecting to ${node} on port ${port}`);
                const timeout = setTimeout(() => {
                    this.disconnect(this.autoReconnect);
                    reject(Error(`timeout`));
                }, this.timeoutConnect);
                socket.on("connect", () => {
                    this.DEBUG_LOG &&
                        console.log(`bsv-p2p: Connected to ${node} on port ${port}`);
                    const payload = messages_1.Version.write({
                        ticker,
                        version,
                        user_agent,
                        start_height,
                        mempoolTxs,
                        segwit: this.segwit,
                        options,
                    });
                    this.sendMessage("version", payload, true);
                    this.emit("connect", { ticker, node, port });
                });
                socket.on("error", (error) => {
                    this.DEBUG_LOG && console.error(`bsv-p2p: Socket error`, error);
                    this.emit("error_socket", { ticker, node, port, error });
                    this.disconnect(this.autoReconnect);
                    clearTimeout(timeout);
                    reject(Error(`disconnected (error)`));
                });
                socket.on("end", () => {
                    this.DEBUG_LOG &&
                        console.warn(`bsv-p2p: Socket disconnected ${node}`);
                    this.disconnect(this.autoReconnect);
                    clearTimeout(timeout);
                    reject(Error(`disconnected (end)`));
                });
                socket.on("data", (data) => {
                    // this.DEBUG_LOG && console.log(`bsv-p2p: data`, data.toString('hex'))
                    try {
                        if (buffers.downloadingBlock) {
                            this.streamBlock(data);
                        }
                        else {
                            buffers.length += data.length;
                            buffers.data.push(data);
                        }
                        if (buffers.length >= buffers.needed) {
                            this.readMessage(Buffer.concat(buffers.data));
                        }
                    }
                    catch (error) {
                        const buffer = Buffer.concat(buffers.data);
                        this.DEBUG_LOG &&
                            console.error(`bsv-p2p: on data error. Disconnecting. buffer: ${buffer.toString("hex")}`, error);
                        const { magic, extmsg } = this;
                        this.emit("error_message", {
                            ticker,
                            node,
                            port,
                            error,
                            magic,
                            extmsg,
                            buffer,
                        });
                        this.disconnect(this.autoReconnect); // TODO: Recover!
                    }
                });
                let connectVrack = false;
                let connectVersion = false;
                const isConnected = () => {
                    if (connectVrack && connectVersion) {
                        clearTimeout(timeout);
                        this.connected = true;
                        resolve();
                        this.emit(`connected`, { ticker, node, port });
                    }
                };
                this.emitter.once("verack", () => {
                    connectVrack = true;
                    isConnected();
                });
                this.emitter.once("version", ({ version }) => {
                    if (this.segwit && !version.segwit) {
                        this.disconnect(false);
                        clearTimeout(timeout);
                        reject(Error(`peer does not support segwit`));
                    }
                    else {
                        connectVersion = true;
                        isConnected();
                    }
                });
                socket.connect(port, node);
            });
        }
        return this.promiseConnect;
    }
    disconnect(autoReconnect = false) {
        this.autoReconnect = !!autoReconnect;
        if (this.socket) {
            this.DEBUG_LOG && console.warn(`bsv-p2p: Disconnected from ${this.node}`);
            this.socket.destroy();
            this.socket = null;
            this.connected = false;
            this.disconnects++;
            this.buffers = {
                data: [],
                needed: 0,
                length: 0,
                chunkNum: 0,
                block: new bsv_minimal_1.Block({ validate: this.validate }),
                downloadingBlock: false,
            };
            delete this.promiseConnect;
            this.emitter.removeAllListeners("disconnected");
            const { ticker, node, port, disconnects } = this;
            this.emit("disconnected", { ticker, node, port, disconnects });
            if (autoReconnect && typeof this.autoReconnectWait === "number") {
                setTimeout(() => this.connect().catch(() => { }), this.autoReconnectWait); // Wait before reconnecting
            }
        }
    }
    getHeaders({ from, to, timeoutSeconds = 60 * 1, }) {
        return __awaiter(this, void 0, void 0, function* () {
            const { version } = this;
            const payload = messages_1.Headers.getheaders({ from, to, version });
            this.sendMessage("getheaders", payload);
            const headers = yield this.emitter.wait("headers", null, timeoutSeconds);
            return headers;
        });
    }
    getMempool() {
        this.sendMessage("mempool", null);
    }
    getBlock(hash, timeoutSeconds) {
        return __awaiter(this, void 0, void 0, function* () {
            if (Buffer.isBuffer(hash)) {
                this.getBlocks([hash]);
                hash = hash.toString("hex");
            }
            else {
                this.getBlocks([Buffer.from(hash, "hex")]);
            }
            if (!timeoutSeconds) {
                timeoutSeconds = this.stream ? 30 : 60 * 10; // Wait at least 30 seconds to start streaming block. 10 minutes otherwise
            }
            const results = yield this.emitter.wait(`block_${hash}`, [`notfound_block_${hash}`, `reject_${hash}`], timeoutSeconds);
            return results;
        });
    }
    getBlocks(blocks) {
        // blocks is an array of 32 byte Buffer block hashes
        if (this.segwit) {
            const payload = messages_1.GetData.write({ witness_blocks: blocks });
            this.sendMessage("getdata", payload);
        }
        else {
            const payload = messages_1.GetData.write({ blocks });
            this.sendMessage("getdata", payload);
        }
    }
    broadcastTx(transaction, timeoutSeconds = 60 * 1) {
        return __awaiter(this, void 0, void 0, function* () {
            const [result] = yield this.broadcastTxs([transaction], timeoutSeconds);
            if (result.status === "rejected") {
                throw Error(result.reason);
            }
            else {
                return result.value;
            }
        });
    }
    broadcastTxs(transactions, timeoutSeconds = 60 * 5) {
        return __awaiter(this, void 0, void 0, function* () {
            if (transactions.length > config_1.MAX_PER_MSG)
                throw Error(`Too many transactions (${config_1.MAX_PER_MSG} max)`);
            const txs = transactions.map((t) => t.getHash());
            const payload = messages_1.Inv.write({ txs });
            this.sendMessage("inv", payload);
            const results = yield Promise.allSettled(transactions.map((tx) => __awaiter(this, void 0, void 0, function* () {
                yield this.emitter.wait(`getdata_tx_${tx.getTxid()}`, `reject_${tx.getTxid()}`, timeoutSeconds);
                this.sendMessage("tx", tx.toBuffer());
            })));
            return results;
        });
    }
    getTxs(txs) {
        if (txs.length === 0)
            return;
        if (this.segwit) {
            const payload = messages_1.GetData.write({ witness_txs: txs });
            this.sendMessage("getdata", payload);
        }
        else {
            const payload = messages_1.GetData.write({ txs });
            this.sendMessage("getdata", payload);
        }
    }
    getAddr(timeoutSeconds = 60 * 2) {
        return __awaiter(this, void 0, void 0, function* () {
            // 2 minute default timeout
            this.sendMessage("getaddr", null);
            const result = yield this.emitter.wait("addr", null, timeoutSeconds);
            return result;
        });
    }
    ping(timeoutSeconds = 30) {
        return __awaiter(this, void 0, void 0, function* () {
            // 30 second default timeout
            const nonce = crypto_1.default.randomBytes(8);
            const id = nonce.toString("hex");
            const date = +new Date();
            this.sendMessage("ping", nonce);
            yield this.emitter.wait(`pong_${id}`, null, timeoutSeconds);
            return +new Date() - date;
        });
    }
    fetchMempoolTxs(filterTxids) {
        if (!this.mempoolTxs)
            throw Error(`mempoolTxs was not set`);
        // Array of announced 32 byte txids from mempool. Return a filtered txid array of the txs you want to download.
        this.listenTxs = filterTxids;
    }
    fetchNewBlocks(filterBlocks) {
        // Array of announced 32 byte block hashes. Return a filtered block array of the blocks you want to download.
        this.listenBlocks = filterBlocks;
    }
}
exports.default = Peer;
//# sourceMappingURL=index.js.map